/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_data.h
//
// This file contains the following declarations 
// 
// C_OLAPComm_Data	:  Virtual abstract class for any arbitrary 
//                         data type for packing and unpacking so it
//                         can send across the wire
// 			    
// C_OLAPComm_Endian_Convertors : Utilities for converting data types
// 			   across various endian types                          
//
// C_OLAPComm_DataLinearizer : Linearized data accumulator for packing 
//                          standard types and derived types from C_OLAPComm_Data
//
// C_OLAPComm_DataDelinearizer : For unpacking a stream of data bytes into 
//                          standard types and derived types from C_OLAPComm_Data 
//  
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_COMM_DATA_H)
#define	__OLAP_COMM_DATA_H	

#ifdef _WINDOWS
#pragma warning(disable:4786)
#include <iostream>
#include <strstream>
using namespace std;
#else
#include <iostream.h>
#include <strstream.h>
#include <string>
#endif

#include "olap_types.h"
#include "olap_comm_types.h"

class C_OLAPComm_DataLinearizer;
class C_OLAPComm_DataDelinearizer;

class C_OLAPComm_Data {

public :

	C_OLAPComm_Data()
	{

	}

	virtual ~C_OLAPComm_Data()
	{

	}

	virtual void Linearize( C_OLAPComm_DataLinearizer &dLin ) = 0 ;

	virtual void DeLinearize( C_OLAPComm_DataDelinearizer &deLin ) = 0 ;

} ;

class C_OLAPComm_Endian_Convertors {

public:

	static void ConvertInt64( int64& iDest, const int64 iSrc )
	{
		ReverseMemory( &iDest, &iSrc, sizeof(int64) ) ;
	}

	static void ConvertInt32( int32& iDest, const int32 iSrc )
	{
		ReverseMemory( &iDest, &iSrc, sizeof(int32) ) ;
	}

	static void ConvertInt16( int16& iDest, const int16 iSrc )
	{
		ReverseMemory( &iDest, &iSrc, sizeof(int16) ) ; 
	}

	static void ConvertFloat( float& fDest, const float fSrc )
	{
		ReverseMemory( &fDest, &fSrc, sizeof(float) ) ;
	}
private:
	
	static void ReverseMemory( void* Destination,
		                       const void*	Source,
							   int32 Length );

} ;

class C_OLAPComm_DataLinearizer {

public :

	C_OLAPComm_DataLinearizer()
	{
		
	}

	void Pack( const int64 iInt64 )
	{
		m_isOutputData.write( (const char *)&iInt64, sizeof(iInt64 ) );
	}

	void Pack( const int32 iInt32 )
	{
		m_isOutputData.write( (const char *)&iInt32, sizeof(iInt32 ) );
	}

	void Pack( const int16 iInt16 )
	{
		m_isOutputData.write( (const char *)&iInt16, sizeof(iInt16 ) );
	}

	void Pack( const int8 iInt8 )
	{
		m_isOutputData.write( (const char *)&iInt8, sizeof(iInt8 ) );
	}

	void Pack( const float fFloat )
	{
		m_isOutputData.write( (const char *)&fFloat, sizeof(fFloat) );
	}

	void Pack( const char* szString )
	{
		int32 iLen = ::strlen(szString) ; 
	
		Pack( iLen ) ;

		m_isOutputData.write( szString, iLen ) ;
	}

	void Pack( const string& str ) 
	{
		Pack( str.c_str() ) ; 
	}

	void Pack( const void* pVoid, int32 size )
	{
		m_isOutputData.write( (const char *)pVoid, size ) ;
	}

	void Pack( C_OLAPComm_Data& isOlapData )
	{
		isOlapData.Linearize( *this ) ; 
	}

	char* GetFreezedLinearizedData()
	{
		return m_isOutputData.str() ;
	}

	char* GetLinearizedData()
	{
		char* szData = m_isOutputData.str() ;

		m_isOutputData.rdbuf()->freeze(0) ;

		return szData ;
	}

	int32 GetLinearizedDataLength()
	{
		return m_isOutputData.pcount() ;
	}
	
private:

	ostrstream		m_isOutputData ;
} ;


class C_OLAPComm_DataDelinearizer {

public :


	C_OLAPComm_DataDelinearizer( char* buf, int32 len,
		OLAPCOMM_ENDIAN_TYPE isEndian = gc_isMyEndian,
		bool free_memory = true) :
		m_isInputData(buf,len), m_isEndianType(isEndian),
		m_bFreeMemory(free_memory)
	{
	}
							
        virtual ~C_OLAPComm_DataDelinearizer()
        {
		if ( m_bFreeMemory )
		{
			m_isInputData.seekg(0);
			delete [] m_isInputData.rdbuf()->str() ; 
		}
        }

	void Unpack( int64& iInt64 )
	{
		m_isInputData.read( (char *)&iInt64, sizeof(iInt64 ) );

		if ( m_isEndianType != gc_isMyEndian )
		{
			int64 iTemp = iInt64 ;
			C_OLAPComm_Endian_Convertors::ConvertInt64( iInt64, iTemp ) ;
		}
	}

	void Unpack( int32& iInt32 )
	{
		m_isInputData.read( (char *)&iInt32, sizeof(iInt32 ) );

		if ( m_isEndianType != gc_isMyEndian )
		{
			int32 iTemp = iInt32 ;
			C_OLAPComm_Endian_Convertors::ConvertInt32( iInt32, iTemp ) ;
		}
	}

	void Unpack( int16& iInt16 )
	{
		m_isInputData.read( (char *)&iInt16, sizeof(iInt16 ) );

		if ( m_isEndianType != gc_isMyEndian )
		{
			int16 iTemp = iInt16 ;
			C_OLAPComm_Endian_Convertors::ConvertInt16( iInt16, iTemp ) ;
		}
	}

	
	void Unpack( int8& iInt8 )
	{
		m_isInputData.read( (char *)&iInt8, sizeof(iInt8 ) );
	}

	void Unpack( float& fFloat )
	{
		m_isInputData.read( (char *)&fFloat, sizeof(fFloat ) );

		if ( m_isEndianType != gc_isMyEndian )
		{
			float fTemp = fFloat ;
			C_OLAPComm_Endian_Convertors::ConvertFloat( fFloat, fTemp ) ;
		}
	}

	void Unpack( char* szString, int32 &iLen ) ;
	
	void Unpack( string& str ) 
	{
		char buf[OLAP_BUFFER_SIZE] ; 
		int32 iBufLen = OLAP_BUFFER_SIZE ; 

		Unpack( buf, iBufLen ) ; 
		str = buf ; 
	}

	void Unpack( void* pVoid, int32 size )
	{
		m_isInputData.read( (char *)pVoid, size ) ;
	}

	void Unpack( C_OLAPComm_Data& isOlapData )
	{
		isOlapData.DeLinearize( *this ) ; 
	}

	int32 GetDeLinearizedDataLength()
	{
		return m_isInputData.gcount() ;
	}

	char* GetFreezedDeLinearizedData()
	{
		return m_isInputData.rdbuf()->str() ;
	}

	char* GetDeLinearizedData()
	{
		char* szData = m_isInputData.rdbuf()->str() ;

		m_isInputData.rdbuf()->freeze(0) ;

		return szData ;
	}

private:

	istrstream		m_isInputData ;
	OLAPCOMM_ENDIAN_TYPE	m_isEndianType ;
	bool			m_bFreeMemory ;
} ;

#endif
