/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_comm_system_shore.cpp
//
// This file contains the implementations of the transport layer for
// shore sockets. The abstracted read/write interface is mapped onto the
// shore sfile read and write for nonblocking
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if defined(SOLARIS2)
#define BSD_COMP
#include <sys/ioctl.h>
#undef BSD_COMP
#endif

#include "olap_types.h"
#include "olap_comm_types.h"
#include "olap_comm_system_shore.h"

#include <arpa/inet.h>

C_OLAPComm_System_Shore::C_OLAPComm_System_Shore()
{
	m_isSocket = new sfile_t ;
	if ( !m_isSocket ) W_FATAL(BORG_OUTOFMEMORY) ;

	m_bAllocated = true ; 
}

C_OLAPComm_System_Shore::C_OLAPComm_System_Shore(const SOCKET fd)
{
	m_isSocket = new sfile_t(fd, sfile_t::OPEN_RDWR) ;
	if ( !m_isSocket ) W_FATAL(BORG_OUTOFMEMORY) ;

	m_bAllocated = true ; 
}

w_rc_t
C_OLAPComm_System_Shore::WillBlockRead( bool& block )
{
	fd_set		rset ; 
	FD_ZERO(&rset) ; 
	FD_SET(m_isSocket->fd(), &rset) ; 
	struct timeval timeout;
	timeout.tv_sec  = 0; timeout.tv_usec = 0;

	int32 n = select( m_isSocket->fd() +1 , &rset, NULL, NULL, &timeout ) ;  

	if ( n < 0 ) W_FATAL( fcOS ) ; 

        block =  n > 0 ? true : false ;

	return RCOK ; 
}

w_rc_t 
C_OLAPComm_System_Shore::Write( void* ptr, int32 len, int32& pkt_write )
{
	char* szNetworkData = (char *)ptr ; 
	int32 temp = 0 ;

        pkt_write = 0 ; 
	while ( pkt_write < len ) 
	{
		if ( m_isSocket->send( szNetworkData + pkt_write, 
				len - pkt_write, 0, temp) != RCOK )
			return RC(BORGCOMM_UNABLE_TO_SEND) ;

		pkt_write += temp ; 
	}

	return RCOK ; 
}

w_rc_t 
C_OLAPComm_System_Shore::Read( void* ptr, int32 len, int32& pkt_read )
{
	char* szNetworkData = (char *)ptr ; 
	int32 temp = 0 ;

        pkt_read = 0 ; 

	while (pkt_read < len)
	{
		if ( m_isSocket->recv( szNetworkData + pkt_read, 
				len - pkt_read, 0, temp) != RCOK ) 
			return RC(BORGCOMM_UNABLE_TO_RECEIVE) ;

		pkt_read += temp;
	}

	return RCOK ; 
}

w_rc_t
C_OLAPComm_System_Shore_Server::Bind( const u_int16 iPort ) 
{
	w_rc_t rc = m_isSocket->socket( AF_INET, SOCK_STREAM, 0 ) ;
	if ( rc ) return RC(BORGCOMM_UNABLE_TO_CREATE_SOCKET);

	struct sockaddr_in		isServAddr ;
	memset( &isServAddr, 0, sizeof(isServAddr) ) ;

	isServAddr.sin_family = AF_INET ;
	isServAddr.sin_port = htons( iPort ) ;
	isServAddr.sin_addr.s_addr = htonl(INADDR_ANY) ;

	rc = m_isSocket->bind( (struct sockaddr *)&isServAddr,
			                   sizeof(isServAddr) ) ; 
	if (rc)	return RC(BORGCOMM_UNABLE_TO_BIND) ;

	rc = m_isSocket->listen( SOMAXCONN ) ; 

	if (rc)	return RC(BORGCOMM_UNABLE_TO_LISTEN) ;

	return RCOK ; 
}

w_rc_t 
C_OLAPComm_System_Shore_Server::Accept( 
				C_OLAPComm_System_Shore_Server& isWorker )
{
	w_rc_t			rc ; 
	struct sockaddr_in	isClientAddr ;
	int32			iAddrLen = sizeof(isClientAddr) ; 

	memset( &isClientAddr, 0, sizeof(isClientAddr) ) ;

	if ( isWorker.m_isSocket )
	{
		rc = isWorker.m_isSocket->close() ;
		if (rc) return RC(BORGCOMM_UNABLE_TO_CLOSE) ; 
	}
	else
	{
		isWorker.m_isSocket = new sfile_t ; 
		if ( !isWorker.m_isSocket ) W_FATAL(BORG_OUTOFMEMORY) ;
	} 

	rc = m_isSocket->accept( (struct sockaddr *)&isClientAddr, &iAddrLen,
				*isWorker.m_isSocket ) ;
	if (rc)	return RC(BORGCOMM_UNABLE_TO_ACCEPT) ;

	return RCOK ;
}

w_rc_t
C_OLAPComm_System_Shore_Client::Connect( const char* szHostName, 
					const u_int16 iPort ) 
{
	if ( m_bConnected ) return RC(BORGCOMM_ALREADY_CONNECTED) ;

	w_rc_t rc = m_isSocket->socket( AF_INET, SOCK_STREAM, 0 ) ;
	if ( rc ) return RC(BORGCOMM_UNABLE_TO_CREATE_SOCKET) ;

	struct hostent*		isServHost ;

	if ( (isServHost = GetHostByName(szHostName)) != NULL )
	{
		struct sockaddr_in		isServAddr ;
		memset( &isServAddr, 0, sizeof(isServAddr) ) ;

		isServAddr.sin_family = AF_INET ;
		isServAddr.sin_port = htons( iPort ) ;

		struct in_addr*  isInAddr = (struct in_addr *) *(isServHost->h_addr_list) ;

		isServAddr.sin_addr.s_addr = inet_addr(inet_ntoa(*isInAddr)) ;

		rc = m_isSocket->connect( (struct sockaddr *)&isServAddr, 
				            sizeof(isServAddr) ) ;
		if (rc)	return RC(BORGCOMM_UNABLE_TO_CONNECT) ;
	}

	return RCOK ; 
}

