/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_operator.h
//
// This file consists of the implementation of some of the methods
// of the abstract operator class.  
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_operator_abstract.h"

C_OLAPOperator_Queue g_OperatorExecQueue ; 

C_OLAPOperator::C_OLAPOperator( C_OLAPPlan_Node* node, int32 numipstreams, int32 numopstreams )
{
	m_isPlanNode = node ; 
	m_iNumInputStreams = numipstreams; 
	m_iNumOutputStreams = numopstreams ; 

	m_isInputStreams = m_isOutputStreams = NULL ; 

	if ( m_iNumInputStreams > 0 )
	{
		m_isInputStreams = new C_OLAPStream* [m_iNumInputStreams] ; 
		if ( !m_isInputStreams ) W_FATAL(BORG_OUTOFMEMORY) ; 

		m_isInputTuples = new C_OLAP_Tuple[m_iNumInputStreams] ; 
		if ( !m_isInputTuples ) W_FATAL(BORG_OUTOFMEMORY) ; 

	}

	if ( m_iNumOutputStreams > 0 )
	{
		m_isOutputStreams = new C_OLAPStream* [m_iNumOutputStreams] ; 
		if ( !m_isOutputStreams ) W_FATAL(BORG_OUTOFMEMORY) ; 

		m_isOutputTuples = new C_OLAP_Tuple[m_iNumOutputStreams] ; 
		if ( !m_isOutputTuples ) W_FATAL(BORG_OUTOFMEMORY) ; 

	}

	m_bFinished = false ; m_bShutdown = false ;  

	// initialize m_isExprInfo
	if (m_isPlanNode && (m_isPlanNode->project_list() || m_isPlanNode->predicate()))
		m_isExprInfo.Init( m_iNumInputStreams, m_iNumOutputStreams,
			   m_isPlanNode->m_isInputTupleTypes, m_isPlanNode->m_isOutputTupleTypes, 
			   m_isInputTuples, m_isOutputTuples,
			   m_isPlanNode->project_list() ? true : false);
}

C_OLAPOperator::~C_OLAPOperator()
{
	if ( m_iNumInputStreams > 0 )
	{
		delete [] m_isInputStreams ;
		m_isInputStreams = NULL ; 
	
		delete [] m_isInputTuples ; 
		m_isInputTuples = NULL ; 

	}

	if ( m_iNumOutputStreams > 0 )
	{
		delete [] m_isOutputStreams ; 
		m_isOutputStreams = NULL ; 

		delete [] m_isOutputTuples ; 
		m_isOutputTuples = NULL ; 

	}
}

w_rc_t C_OLAPOperator::WaitForCompletion()
{
	W_DO( m_isMutex.acquire() ) ; 

	while ( !m_bFinished )
		W_DO( m_isCond.wait( m_isMutex ) ) ; 

	m_isMutex.release() ; 

	return RCOK ; 
}

w_rc_t C_OLAPOperator::OperatorFinished()
{
	W_DO( m_isMutex.acquire() ) ; 
	m_bFinished = true ; 
	m_isMutex.release() ; 

	m_isCond.signal() ;
	return RCOK ; 
}

void C_OLAPOperator::PropagateException( C_OLAPStream* exp_stream )
{
	cerr << "Number of input streams " << m_iNumInputStreams << endl ; 

	for ( int32 i = 0 ; i < m_iNumInputStreams ; i++ )
	{
		if ( m_isInputStreams[i] && 
		     (m_isInputStreams[i] != exp_stream) ) 
                    // necessary since for some operators the input streams
		    // are set up later
		    W_IGNORE( m_isInputStreams[i]->SendExceptionDownStream() ) ; 
	}

	for ( int32 i = 0 ; i < m_iNumOutputStreams ; i++ )
	{
		if ( m_isOutputStreams[i] &&
		     (m_isOutputStreams[i] != exp_stream) ) 
		    W_IGNORE( m_isOutputStreams[i]->SendExceptionUpStream() ) ; 
	}
}

C_OLAPOperator_Queue::~C_OLAPOperator_Queue()
{
	C_OLAPOperator*	request = NULL ; 

	while ( (request = m_isRequestQueue.pop()) )
		delete request ; 
} 

w_rc_t 
C_OLAPOperator_Queue::enqueue( C_OLAPOperator* request )
{
	W_DO( m_isQueueMutex.acquire() ) ; 

	while ( m_isRequestQueue.num_members() >= max_requests ) {
		W_COERCE( m_isQueueCond.wait( m_isQueueMutex ) ) ;
	}	

	m_isRequestQueue.append( request ) ; 

	m_isQueueMutex.release() ; 
	m_isQueueCond.signal() ; 

	return RCOK ; 
}

w_rc_t C_OLAPOperator_Queue::dequeue( C_OLAPOperator*& request )
{
	W_DO( m_isQueueMutex.acquire() ) ; 

	while ( m_isRequestQueue.num_members() <= 0 ) {
		W_COERCE( m_isQueueCond.wait( m_isQueueMutex ) ) ;
	}

	request = m_isRequestQueue.pop() ; 

	m_isQueueMutex.release() ; 
	m_isQueueCond.signal() ; 

	return RCOK ; 
}

void C_OLAPOperator_Thread::run()
{
	C_OLAPOperator*		op_request = NULL ; 

	while (1)
	{
		W_COERCE( g_OperatorExecQueue.dequeue( op_request ) ) ;  

		if ( op_request->IsShutdown() ) break ; 
			
		W_IGNORE( op_request->start() ) ; 

		op_request->GetRC() = op_request->run() ; 

		if ( op_request->GetRC() && 
		     op_request->GetRC().err_num() != 
					BORGOP_ABNORMAL_TERMINATION )
		{
			op_request->PropagateException() ; 
		}

		W_IGNORE( op_request->shutdown() ) ; 

		W_IGNORE( op_request->OperatorFinished() ) ; 
	}
}

#ifdef __GNUG__
template class w_list_t<C_OLAPOperator>;
template class w_list_i<C_OLAPOperator>;
#endif
