/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_operator_abstract.h
//
// This file consists of the abstract definition of operator class. 
// Any class derived from it requires that it implements the start,
// run and shutdown methods. 
//
// Also consists of the definitions of the operator queue and operator
// thread.
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_OPERATOR_ABSTRACT_H)
#define	__OLAP_OPERATOR_ABSTRACT_H

#include "olap_server_globals.h"

#include "w_list.h"

#include "olap_borg_thread.h"
#include "olap_stream.h" 
#include "borgop_error_def_gen.h"

#include "olap_predicate.h"        // For predicate processing...
#include "olap_expr_abstract.h"    // For ExprInfo, ExprValue
#include "olap_attr.h"             // ProjectList

#include "olap_plan_node.h"

#define	W_EXCEPTION( stream, x )			\
{							\
    if ((x))						\
    {							\
	this->PropagateException( (stream) ) ;		\
	cerr << "Operator shutting down" << endl ;	\
	return RC(BORGOP_ABNORMAL_TERMINATION) ; 	\
    }							\
}	

class C_OLAPPlan_Node ; 

class C_OLAPOperator 
{
public:

	C_OLAPOperator(	C_OLAPPlan_Node* node, int32 numipstreams, int32 numopstreams )  ;

	virtual ~C_OLAPOperator() ; 

	virtual w_rc_t start() = 0 ; 
	virtual w_rc_t run() = 0 ; 
	virtual w_rc_t shutdown() = 0 ; 

	bool DidErrorOccurred() 
	{
		return m_isRc ? true : false ;
	}

	w_rc_t& GetRC()
	{
		return m_isRc ;  
	}

	w_rc_t WaitForCompletion() ; 

	bool IsLastOperator()
	{
		return m_bLastOp ; 
	}

	bool IsShutdown()
	{
		return m_bShutdown ; 
	}

protected:

	void LastOperator()
	{
		m_bLastOp = true ; 
	}

	void SetRC( w_rc_t& rc )
	{
		m_isRc = rc ; 
	}

	C_OLAPStream* InputStream(int32 idx)
	{
		return m_isInputStreams[idx] ; 
	}

	C_OLAPStream* OutputStream(int32 idx)
	{
		return m_isOutputStreams[idx] ; 
	}

	void SetInputStream( int32 idx, C_OLAPStream* anInputStream ) 
	{
		m_isInputStreams[idx] = anInputStream ; 
	}

	void SetOutputStream( int32 idx, C_OLAPStream* anOutputStream ) 
	{
		m_isOutputStreams[idx] = anOutputStream ; 
	}	

        void PropagateException( C_OLAPStream* exp_stream = NULL ) ;

	w_rc_t OperatorFinished() ;

	/* ---------------------------------------------------- */
	// Processing of expressions and Projection

	bool Evaluate()
	{
		if ( m_isPlanNode->predicate() ) 
			return m_isPlanNode->predicate()->evaluate(&m_isExprInfo);

		return true;    // No condition
	}

	bool ProjectTuple()
	{
		ExprValue ev;
		if ( m_isPlanNode->project_list() ) 
			return m_isPlanNode->project_list()->Eval(m_isExprInfo, ev);

		// Otherwise output tuple points to the input tuple
		return true;
	}

	ExprInfo            m_isExprInfo;

	int32 			m_iNumInputStreams ; 
	int32 			m_iNumOutputStreams ; 
	C_OLAPStream**  	m_isInputStreams ; 
	C_OLAPStream**		m_isOutputStreams ;  

	// Used in the run method of every operator
	C_OLAP_Tuple*       	m_isInputTuples ;	
	C_OLAP_Tuple*       	m_isOutputTuples ;	

	C_OLAPPlan_Node*	m_isPlanNode ; 

	w_rc_t		m_isRc ; 
	w_link_t	m_isLink ; 
	bool		m_bLastOp ; 

	bool		m_bFinished ;
	smutex_t	m_isMutex ;
	scond_t		m_isCond ;

	bool		m_bShutdown ; 

	friend class C_OLAPOperator_Queue ; 
	friend class C_OLAPOperator_Thread ; 
	friend class C_OLAP_Physical_Plan ; 
} ; 

class C_OLAPOperator_Queue
{
public:

	enum {
		max_requests = 256
	} ; 

	C_OLAPOperator_Queue() : m_isRequestQueue(
					offsetof(C_OLAPOperator, m_isLink)) 
	{ } 

	~C_OLAPOperator_Queue() ;

	w_rc_t enqueue( C_OLAPOperator* request ) ;
	w_rc_t dequeue( C_OLAPOperator*& request ) ;  

protected:

	w_list_t<C_OLAPOperator>	m_isRequestQueue ;	

	scond_t				m_isQueueCond ; 
	smutex_t			m_isQueueMutex ; 	

private:
} ;

class C_OLAPOperator_Thread : public C_OLAPBorg_Thread
{
public :

	C_OLAPOperator_Thread() 
	{ }

	~C_OLAPOperator_Thread()
	{ } 

	virtual void run() ;

	virtual olap_thread_type GetThreadType()
	{
		return OLAP_OPERATOR_THREAD ;
	}

protected :

} ; 

#endif
