/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_expr.cpp
//
// This file contains the implementation for expressions.
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////

#include "olap_expr.h"
#include <stdio.h>    // sprintf
#include <stdlib.h>   // atoi etc...


/* ----------------- Eval: C_OLAPLogicalExpr ------------------- */

bool C_OLAPLogicalExpr::Eval(ExprInfo& einfo, ExprValue& expval)
{
    // Make AND, ORs shortcircuit!!
    // a NOT has only one argument!
    ExprValue NewEV;
    bool bRet = arg(0)->Eval(einfo, expval);
    if (!bRet) return false;

    if (_letype == Not)
    {
        expval.b = ! expval.b;
        return true;
    }

    for (int i = 1; i < nargs(); i++)
    {
        bRet = arg(i)->Eval(einfo, NewEV);
        if (!bRet) return false;
        switch(_letype)
        {
        case And:
            expval.b = expval.b && NewEV.b;
            if (expval.b == false) return true;  // Short circuit...
            break;
        case Or:
            expval.b = expval.b || NewEV.b;
            if (expval.b == true) return true;  // Short circuit...
            break;
        default:
            assert(0);
        }
    } // for

    return true;
}

void C_OLAPLogicalExpr::WriteTree(ostream& fs)
{
  C_OLAPExprNode::WriteTree(fs);
  fs << "(C_OLAPLogicalExpr (type = " << int(_letype) << ") ";
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->WriteTree(fs);
  
  fs << ")\n";
}

void C_OLAPLogicalExpr::ReadTree(istream& fs)
{
  C_OLAPExprNode::ReadTree(fs);
  char buf[32];
  int32 tint;
  fs >> buf >> buf >> buf
     >> tint >> buf;
  _letype = letype_t(tint);
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->ReadTree(fs);
  
  fs >> buf;
}



/* ----------------- Eval: C_OLAPArithExpr ------------------- */

bool C_OLAPArithExpr::Eval(ExprInfo& einfo, ExprValue& expval)
{
     // The input expval is used for accumumating the quantity and EvalEV
    // computes individual branches...
    ExprValue EvalEV;
    arg(0)->Eval(einfo, expval);
    if (_aetype == UnaryPlus)
        return true;
    else if (_aetype == UnaryMinus)
    {
        switch (expval.type) {
        case BORG_INT:   expval.i = -expval.i;  break;
        case BORG_FLOAT: expval.f = -expval.f;  break;
        default: return false; // failed!!
        }
        return true;
    }

    for (int i = 1; i < nargs(); i++)
    {
        arg(i)->Eval(einfo, EvalEV);
        if (expval.type != EvalEV.type)
            MakeSameType(expval, EvalEV);

        switch(_aetype)
        {
        case Plus:
            if (ProcessPlus(expval, EvalEV) == false) return false;
            break;
        case Minus:
            if (!ProcessMinus(expval, EvalEV)) return false;
            break;
        case Mul:
            if (!ProcessMul(expval, EvalEV)) return false;
            break;
        case Div:
            if (!ProcessDiv(expval, EvalEV)) return false;
            break;
        default: assert(0);
        }
    }
    return true;
}


void C_OLAPArithExpr::WriteTree(ostream& fs)
{
  C_OLAPExprNode::WriteTree(fs);
  fs << "(C_OLAPArithExpr (type = " << int(_aetype) << ") ";
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->WriteTree(fs);
  
  fs << ")\n";
}

void C_OLAPArithExpr::ReadTree(istream& fs)
{
  C_OLAPExprNode::ReadTree(fs);
  char buf[32];
  int32 tint;
  fs >> buf >> buf >> buf
     >> tint >> buf;
  _aetype = aetype_t(tint);
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->ReadTree(fs);
  
  fs >> buf;
}



/* ----------------- Eval: C_OLAPCompExpr -------------------- */

bool C_OLAPCompExpr::Eval(ExprInfo& einfo, ExprValue& expval)
{
    ////Take care of In

    expval.type = BORG_UNDEF;     // actually
    if (_cetype >= Eq && _cetype <= Le)
    {
        ExprValue Lev, Rev;
        arg(0)->Eval(einfo, Lev);
        arg(1)->Eval(einfo, Rev);

        // Get both arguments to be of the same type
        int comp_code = CompareExprValue(Lev, Rev);
        bool ret_value = false;
        switch (_cetype)
        {
        case Eq:
            if (comp_code == 0) 
                ret_value = true;
            break;
        case Gt:
            if (comp_code == 1) 
                ret_value = true;
            break;
        case Lt:
            if (comp_code == -1) 
                ret_value = true;
            break;
        case Ge:
            if (comp_code != -1) 
                ret_value = true;
            break;
        case Le:
            if (comp_code != 1) 
                ret_value = true;
            break;
        default:
            assert(0);
	}
	expval.b = ret_value;
	return true;
    }
    else if (_cetype == Between)
    {
        ExprValue Lev, Rev, R2ev;
        arg(0)->Eval(einfo, Lev);
        arg(1)->Eval(einfo, Rev);
        arg(1)->Eval(einfo, R2ev);

        // Get both arguments to be of the same type
        int comp_code  = CompareExprValue(Lev, Rev);
        int comp_code2 = CompareExprValue(Lev, R2ev);
        if (comp_code == 1 && comp_code2 == -1)
            expval.b = true;
        else
            expval.b = false;

        return true;
    }

    return false;
}



void C_OLAPCompExpr::WriteTree(ostream& fs)
{
  C_OLAPExprNode::WriteTree(fs);
  fs << "(C_OLAPCompExpr (type = " << int(_cetype) << ") ";
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->WriteTree(fs);
  
  fs << ")\n";
}


void C_OLAPCompExpr::ReadTree(istream& fs)
{
  C_OLAPExprNode::ReadTree(fs);
  char buf[32]; int32 tint;
  fs >> buf >> buf >> buf
     >> tint >> buf;
  _cetype = cetype_t(tint);
  
  for (int i = 0; i < nargs(); i++)
    arg(i)->ReadTree(fs);
  
  fs >> buf;
}
