/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_expr.h
//
// This file contains the interface defintion for expressions and
// their evaluation in terms of tuples. 
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_EXPR_H)
#define __OLAP_EXPR_H

#include "olap_expr_abstract.h"

/* -------------------- C_OLAPLogicalExpr ----------------------- */

class C_OLAPLogicalExpr : public C_OLAPExprNode {
public:
    enum letype_t {
	And = 0, Or, Not, LE_Invalid
    };

private:
    // Additional information that is stored for this node
    letype_t _letype;

    /* ---------------------------------------------------------- */
    // These function are used to process measures
public:
    C_OLAPLogicalExpr(letype_t t) : C_OLAPExprNode(LogicalExpr) 
    { _letype = t; }

    C_OLAPLogicalExpr() : C_OLAPExprNode(LogicalExpr) 
    { _letype = LE_Invalid; }

    letype_t& letype() { return _letype; }

    void AddLogicalExpr(pC_OLAPExprNode pqn) { AddArg(pqn); }

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    // It returns false on error...
    bool Eval(ExprInfo& einfo, ExprValue& expval);

    /* ------------------------------------------------------ */
    // Function which dumps the tree to/from a file 'fs'
    void WriteTree(ostream& fs);
    void ReadTree(istream& fs);
};



/* ---------------------- C_OLAPArithExpr ------------------------ */

// An ArithExpr contains arithmetic expressions
// children can be ArithExpr, Attribute, Value

class C_OLAPArithExpr : public C_OLAPExprNode {
public:
    enum aetype_t {
	Plus = 0, Minus, Mul, Div, UnaryPlus, UnaryMinus, AE_Invalid
    };

private:
    // Additional information that is stored for this node
    aetype_t _aetype;

    /* ----------------------------------------------------------- */
    // These function are used to process measures
public:
    C_OLAPArithExpr(aetype_t t) : C_OLAPExprNode(ArithExpr) 
    { _aetype = t; }

    C_OLAPArithExpr() : C_OLAPExprNode(ArithExpr) 
    { _aetype = AE_Invalid; }

    aetype_t& aetype() { return _aetype; }

    void AddArithExpr(pC_OLAPExprNode pqn) { AddArg(pqn); }

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    // It returns false on error...
    bool Eval(ExprInfo& einfo, ExprValue& expval);

    /* ------------------------------------------------------ */
    // Function which dumps the tree to/from a file 'fs'
    void WriteTree(ostream& fs);
    void ReadTree(istream& fs);

protected:
    void MakeSameType(ExprValue& v1,  ExprValue& v2);
    bool ProcessPlus(ExprValue& Acc,  ExprValue& v);
    bool ProcessMinus(ExprValue& Acc, ExprValue& v);
    bool ProcessMul(ExprValue& Acc,   ExprValue& v);
    bool ProcessDiv(ExprValue& Acc,   ExprValue& v);
};



/* ---------------------- C_OLAPCompExpr ------------------------- */

// A CompExpr contains comparisons
// Children can be Attributes, or values

class C_OLAPCompExpr : public C_OLAPExprNode {
public:
    enum cetype_t {
	Eq = 0, Gt, Ge, Lt, Le, Between, CE_Invalid
    };

private:
    // Additional information that is stored for this node
    cetype_t _cetype;

    /* ----------------------------------------------------------- */
    // These function are used to process measures
public:
    C_OLAPCompExpr(cetype_t t) : C_OLAPExprNode(CompExpr) 
    { _cetype = t; }

    C_OLAPCompExpr() : C_OLAPExprNode(CompExpr) 
    { _cetype = CE_Invalid; }

    cetype_t& cetype() { return _cetype; }

    void AddCompExpr(pC_OLAPExprNode pqn) { AddArg(pqn); }

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    // It returns false on error...
    bool Eval(ExprInfo& einfo, ExprValue& expval);
    int CompareExprValue(ExprValue& expval1, ExprValue& expval2);

    /* ------------------------------------------------------ */
    // Function which dumps the tree to/from a file 'fs'
    void WriteTree(ostream& fs);
    void ReadTree(istream& fs);
};

/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */


inline void C_OLAPArithExpr::MakeSameType(ExprValue& v1, ExprValue& v2)
{
    if (v1.type == BORG_INT)
    {
        v1.type = BORG_FLOAT;
        v1.f = float(v1.i);
    }
    else if (v2.type == BORG_INT)
    {
        v2.type = BORG_FLOAT;
        v2.f = float(v2.i);
    }
    else assert(0);
    return;
}

/* ----------------------------------------------------------- */
inline bool C_OLAPArithExpr::ProcessPlus(ExprValue& Acc, ExprValue& v)
/* ----------------------------------------------------------- */
{
    if (Acc.type == BORG_INT)
        Acc.i += v.i;
    else
        Acc.f += v.f;

     return true;
}

inline bool C_OLAPArithExpr::ProcessMinus(ExprValue& Acc, ExprValue& v)
{
    if (Acc.type == BORG_INT)
        Acc.i -= v.i;
    else
        Acc.f -= v.f;

     return true;
}

/* ----------------------------------------------------------- */
inline bool C_OLAPArithExpr::ProcessMul(ExprValue& Acc, ExprValue& v)
{
    if (Acc.type == BORG_INT)
        Acc.i *= v.i;
    else
        Acc.f *= v.f;

     return true;
}

/* ----------------------------------------------------------- */
inline bool C_OLAPArithExpr::ProcessDiv(ExprValue& Acc, ExprValue& v)
{
    if (Acc.type == BORG_INT)
    {
        assert(v.i != 0);
        Acc.i /= v.i;
    }
    else
    {
        assert(v.f != 0);
        Acc.f /= v.f;
    }

     return true;
}


/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */
// 1  => Left >  Right
// 0  => Left == Right
// -1 => Left <  Right

inline int C_OLAPCompExpr::CompareExprValue(ExprValue& expval1, 
					    ExprValue& expval2)
{
    if (expval1.type == expval2.type)
    {
	switch (expval1.type) 
	{
	case BORG_INT:
	    if (expval1.i == expval2.i) return 0;
	    if (expval1.i > expval2.i)  return 1;
	    return -1;

	case BORG_FLOAT:
	    if (expval1.f == expval2.f) return 0;
	    if (expval1.f > expval2.f)  return 1;
	    return -1;
	    
	case BORG_STR:
	    return strcmp(expval1.buf, expval2.buf);

	default:
	    assert(0);
	}
    }

    // Comparison between str and non-strings not allowed
    assert(expval1.type != BORG_STR && expval2.type != BORG_STR);
    float v1, v2;

    if (expval1.type == BORG_INT)
    {
        v1 = float(expval1.i);
        v2 = expval2.f;
    }
    else
    {
        v1 = expval1.f;
        v2 = float(expval2.i);
    }

    if (v1 == v2) return 0;
    if (v1 > v2)  return 1;
    return -1;
}

#endif
