/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_expr_attr.h
//
// This file contains the abstract base class for expressions.
//
// Author: Amit Shukla
//
/////////////////////////////////////////////////////////////////////

#if !defined(__OLAP_EXPR_ABSTRACT_H)
#define __OLAP_EXPR_ABSTRACT_H

/* -------------------------------------------------------------- */
// Functions that will be shared by operators. This includes
// a Column Identifier,

#include "w_dynarray.h"
#include "borgtypes.h"
#include <string>

#include "olap_tuple.h"
#include "olap_predicate.h"

/* --------------------- ColumnIdentifier ----------------------- */ 

// Identifies a column uniquely
struct ColumnIdentifier
{
    enum StreamID {left = 0, right = 1, output = 2};

    StreamID strm;
    int colNum;

    ColumnIdentifier& operator=(ColumnIdentifier& ci)
    { strm = ci.strm; colNum = ci.colNum; return *this; }
};


// This variable contains all the information required for
// evaluating an expression on tuples
struct ExprInfo                // [2] is used for the output tuple
{   // 0 -> left, 1 -> right
    C_OLAP_TupleInfo *tInfo[3];
    C_OLAP_Tuple     *tup[3];

    void Init( u_int32 numInputStreams, u_int32 numOutputStreams,
			    TupleTypes &ipTypes, TupleTypes &opTypes,
			    C_OLAP_Tuple *ipTuples, C_OLAP_Tuple *opTuples,
			    bool isProject);
};

// The temporary value generated during a computation
struct ExprValue
{
    union {
	bool  b;                  // Returns boolean values
	int   i;                  // Returns integer values stores length for strings
	float f;                  // Returns float   values
    };
    char  buf[MAX_NAME_SIZE]; // for strings (currently unused??)
    BorgTypes type;
};


/* -------------------------------------------------------------- */ 
/* -- The  base class for the querytree:                       -- */
/* -- C_OLAPExprNode, which supports an N-ary tree.            -- */
/* -------------------------------------------------------------- */ 
/* -------------------------------------------------------------- */ 
/* -- There are three kinds of expressions:                    -- */
/* --  o C_OLAPLogicalExpr - a logical expression (AND, OR)    -- */
/* --  o C_OLAPArithExpr   - arithmetic expression (Sales*1.5) -- */
/* --  o C_OLAPCompExpr    - a comparison expression (<, >)    -- */
/* -------------------------------------------------------------- */ 

// A logical expression contains arguments that are either
// Comparisons or other logical expressions

#include <assert.h>

#ifdef _WINDOWS
#include <fstream>
#else
#include <fstream.h>
class ostrstream;
class ostream;
#endif


/* -------------------------------------------------------------- */ 

class C_OLAPExprNode;
typedef C_OLAPExprNode* pC_OLAPExprNode;

class C_OLAPLogicalExpr;
typedef C_OLAPLogicalExpr* pC_OLAPLogicalExpr;

class C_OLAPArithExpr;
typedef C_OLAPArithExpr* pC_OLAPArithExpr;

class C_OLAPCompExpr;
typedef C_OLAPCompExpr* pC_OLAPCompExpr;

class C_OLAPAttr;
typedef C_OLAPAttr* pC_OLAPAttr;

class C_OLAPProjectList;
typedef C_OLAPProjectList* pC_OLAPProjectList;

class C_OLAPValue;
typedef C_OLAPValue* pC_OLAPValue;

/* ---------------------- C_OLAPExprNode ------------------------ */ 

class C_OLAP_GeneralExpr : public C_OLAP_Predicate
{
    C_OLAPExprNode* pen;
public:
    C_OLAP_GeneralExpr(C_OLAPExprNode* _pen) : pen(_pen)  { }
    virtual ~C_OLAP_GeneralExpr();
    virtual bool evaluate(ExprInfo* pei);
};


class C_OLAPExprNode {
public:

    enum opcode_t {
        LogicalExpr, ArithExpr, CompExpr,
	Attribute, ProjList, Value
    }; 

    class darray {      // Should be used only for append
    private:
	pC_OLAPExprNode* _arr;
	int _size;
	enum { increment = 5 };
    public:
	darray(int size = 10)
	{ _size = size; _arr = new pC_OLAPExprNode [_size];
          for (int i = 0; i < _size; i++) _arr[i] = NULL; }

        ~darray() { delete [] _arr; }

	int size(void) const { return _size; }

	void set_size(int new_size);
	int enlarge(void);

	pC_OLAPExprNode& operator[](int i)
	{ return _arr[i]; }
    };

private:
    opcode_t _opcode;
    int      _nargs;
    darray   _args;

public:
    C_OLAPExprNode(opcode_t t) { _opcode = t; _nargs = 0; }
    virtual ~C_OLAPExprNode()
    { for (int i = 0; i < nargs(); i++)  delete arg(i); }

    opcode_t& opcode(void)      { return _opcode;  }
    int nargs(void) const       { return _nargs;   }
    pC_OLAPExprNode& arg(int n) { return _args[n]; }

    bool AddArg(pC_OLAPExprNode pqn);
    bool RemoveArg(int argnum);

    /* ------------------------------------------------------ */
    // The eval function - has to implemented by all classes
    // It returns false on error...
    virtual bool Eval(ExprInfo& einfo, ExprValue& expval) = 0;

    /* ------------------------------------------------------ */
    // Functions to dump the tree to/from a file 'fs'
    virtual void WriteTree(ostream& fs);
    virtual void ReadTree(istream& fs);
};

#endif
