/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

//////////////////////////////////////////////////////////////////
//
// olap_client_thread.cpp
//
// This file defines the implementation of client service and worker
// threads. 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "w_stream.h"

#include "olap_server_macros.h"
#include "olap_server_globals.h"
#include "olap_bkend_request.h"
#include "olap_client_thread.h"

#include "olap_stream.h"
#include "olap_operator.h"
#include "olap_pplan.h"

extern C_OLAP_Clients		g_Clients ;
extern C_OLAPOperator_Queue	g_OperatorExecQueue ;

C_OLAP_Clients::~C_OLAP_Clients()
{
	W_COERCE( m_isMutex.acquire() ) ; 

	CListElem*			client = NULL ; 

	while ( (client = m_isList.pop()) ) 
		delete client ; 

	m_isMutex.release() ; 
}

rc_t C_OLAP_Clients::insert( const C_OLAPClient_Id& client_id,
			     C_OLAPClient_Thread* client_thread )
{
	CListElem*	new_client = new CListElem ; 

	if ( !new_client ) return RC(BORG_OUTOFMEMORY) ; 	

	new_client->m_isClientId = client_id ; 
	new_client->m_isThread = client_thread ; 
	new_client->m_bActive = client_id ; 

	W_DO( m_isMutex.acquire() ) ; 
	m_isList.append( new_client ) ; 
	m_isMutex.release() ; 	

	return RCOK ; 
}

rc_t C_OLAP_Clients::remove( const C_OLAPClient_Id& client_id )
{
	W_DO( m_isMutex.acquire() ) ; 

	w_list_i<CListElem>		iter(m_isList) ;		
	CListElem*			client = NULL ; 

	while ( (client = iter.next()) )
	{
		if ( client_id == client->m_isClientId ) 
		{
			client->m_bActive = false ;  
			break ; 
		}
	}

	m_isMutex.release() ; 	
	return RCOK ; 
}

rc_t C_OLAP_Clients::GetActiveClients( C_OLAPComm_DataLinearizer& lin )
{
	w_list_i<CListElem>		iter(m_isList) ;		
	CListElem*			client = NULL ; 

	W_DO( m_isMutex.acquire() ) ; 

	lin.Pack( int32(m_isList.num_members()) ) ; 

	while ( (client = iter.next()) )
	{
		if ( client->m_bActive ) 
			client->m_isClientId.Linearize( lin ) ; 
	}
	
	m_isMutex.release() ; 	
	return RCOK ; 
}

rc_t C_OLAP_Clients::RemoveInactiveClients()
{
	CListElem*			client = NULL ; 
	u_int32				num_members;

	W_DO( m_isMutex.acquire() ) ; 
	num_members = m_isList.num_members();

	for ( u_int32 i = 0 ; i < num_members ; i++ )
	{
		client = m_isList.pop() ; 	

		if ( !client->m_bActive )
		{
			delete client->m_isThread ; 
			delete client ; 
		}
		else 
			m_isList.append( client ) ; 
	}
	
	m_isMutex.release() ; 	

	return RCOK ; 
}

rc_t C_OLAP_Clients::WaitForActiveClientsFinish()
{
	// We should not put mutexes here since it will
	// result in deadlock

	w_list_i<CListElem>		iter(m_isList) ;		
	CListElem*			client = NULL ; 

	while ( (client = iter.next()) )
	{
		if ( client->m_bActive ) 
			W_COERCE( client->m_isThread->wait( C_OLAPClient_Thread::WAIT_FOREVER ) ) ; 
	}
	
	return RCOK ; 
}

void C_OLAPClient_Service_Thread::run()
{
	C_OLAPClient_Thread*	clthread = NULL ;

	while (1)
	{
		struct sockaddr_in client_addr ;
		int client_addr_len = sizeof(client_addr) ;
		memset( &client_addr, 0, client_addr_len ) ;

		sfile_t*           client_sock = new sfile_t ;
		if ( !client_sock ) W_FATAL(BORG_OUTOFMEMORY) ; 

		// Now clean up inactive clients
		W_COERCE( g_Clients.RemoveInactiveClients() ) ; 

		// Block until a connection comes in 
		w_rc_t rc = m_isListener->accept( 
				(struct sockaddr *)&client_addr,
				&client_addr_len, *client_sock );

		// Check error in listening
		if ( rc )	
		{
			delete client_sock ; break ; 
		} 
			
		C_OLAPClient_Id		clid( client_sock->fd() ) ; 

		// Create a new thread for the client
		clthread = new C_OLAPClient_Thread( client_sock, clid )  ;  
		if ( !clthread ) W_FATAL(BORG_OUTOFMEMORY) ; 

		// Add an entry to active list client
		W_COERCE( g_Clients.insert( 
				C_OLAPClient_Id(client_sock->fd()),
				clthread ) ) ;

		// Start the thread
		W_COERCE( clthread->fork() ) ;
	}
}

C_OLAPClient_Thread::~C_OLAPClient_Thread()
{
	W_IGNORE( m_isClientSock->close() ) ; 
	delete m_isClientSock ; 
}

void C_OLAPClient_Thread::run()
{
	C_OLAPComm_System_Shore			isServer(m_isClientSock) ; 
	C_OLAPComm_Reconciler			isReconciler ; 
	C_OLAPComm_Packet*			isPkt = NULL ; 

	C_OLAPTuple_ShoreClientStream*		result = NULL ;

	result = new C_OLAPTuple_ShoreClientStream(m_isClientSock) ;
	if ( !result ) W_FATAL(BORG_OUTOFMEMORY) ;

	w_auto_delete_t<C_OLAPTuple_ShoreClientStream>	isStreamDel(result) ; 

	while ( isServer.ReceiveData( isReconciler, isPkt ) == RCOK  )
	{
	   {
		cerr << "Received a new request " << m_isClientSock->fd() << endl ; 
                C_OLAPComm_DataDelinearizer  *qData;
                C_OLAPComm_Generic_Packet    *recvPkt ;

                recvPkt = (C_OLAPComm_Generic_Packet *) isPkt ;
                qData = recvPkt->GetDataDelinearizer();

                // Extract the query tree...here

		delete isPkt ; 

		C_OLAP_Physical_Plan		pplan ; 

		// choose the test plan you want to run here... 
	
                C_OLAP_TupleInfo* tupInfo = pplan.GetRoot()->op_tuple_type(0);

		C_OLAPComm_DataLinearizer isData ;

                // Linearize the metadata
                isData.Pack(*tupInfo);

                C_OLAPComm_Generic_Packet isPacket(isData) ;
                cout << "SERVER: Sending Query Results... " << endl;

                if ( isServer.SendData( isReconciler, isPacket ) <= 0 )
                {
                    cerr << "Unable to send data..." << endl ;
                    ::exit(1) ;
                }


		W_IGNORE( pplan.executePlan( result ) ) ; 

		// W_IGNORE( result->SendEndOfStream() );

		// Reset the result stream
		result->Reset() ; 
            }
	}

	cerr << "Client terminated... " << endl ; 

	W_COERCE( g_Clients.remove( m_isClientId ) ) ;
}

#ifdef __GNUG__
template class w_list_t<C_OLAP_Clients::CListElem>;
template class w_list_i<C_OLAP_Clients::CListElem>;
#endif
