/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

///////////////////////////////////////////////////////////////////
//
// olap_tuple_memory_stream.cpp
//
// This file defines the memory tuple stream for moving tuples from
// one operator to another operator 
//
// Author: Karthikeyan Ramasamy
//
/////////////////////////////////////////////////////////////////////

#include "olap_server_globals.h"
#include "olap_tuple_memory_stream.h"

C_OLAPTuple_MemoryStream::mem_stream_t::mem_stream_t()
{
	m_isPage = new C_OLAP_Page ; 
	if ( !m_isPage ) W_FATAL( BORG_OUTOFMEMORY ) ; 
}

C_OLAPTuple_MemoryStream::mem_stream_t::~mem_stream_t()
{
	if ( m_isPage ) delete m_isPage ;  
}

void C_OLAPTuple_MemoryStream::init()
{
	//cerr << "creating a memory stream" << endl ; 

	m_currInput = m_currOutput = NULL ; 

	for ( int i = 0 ; i < m_iBuffers ; i++ )
	{
		mem_stream_t*	isMemStream = new mem_stream_t ;
		if ( !isMemStream ) W_FATAL(BORG_OUTOFMEMORY) ; 

		//cerr << "Allocated mem page " << isMemStream << endl ; 
		isMemStream->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
		isMemStream->m_isPage->reset() ; 

		m_isInputList.push( isMemStream ) ; 
	}

	m_isInputException = new mem_stream_t ;
	if ( !m_isInputException ) W_FATAL(BORG_OUTOFMEMORY) ; 

	m_isOutputException = new mem_stream_t ;
	if ( !m_isOutputException ) W_FATAL(BORG_OUTOFMEMORY) ; 

	m_isInputException->m_isMsgType = OLAP_ABNORMAL_TERMINATION ; 
	m_isOutputException->m_isMsgType = OLAP_ABNORMAL_TERMINATION ; 
}

void C_OLAPTuple_MemoryStream::Reset()
{
	m_currInput = m_currOutput = NULL ; 
	mem_stream_t*	isMemStream = NULL ;

	while ( ( isMemStream = m_isOutputList.pop() ) ) 
	{
		if ( isMemStream != m_isInputException && 
                     isMemStream != m_isOutputException )
		{
			isMemStream->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
			isMemStream->m_isPage->reset() ; 
			m_isInputList.append( isMemStream ) ; 
		}
	}

	m_isInputException->m_isMsgType = OLAP_ABNORMAL_TERMINATION ; 
	m_isOutputException->m_isMsgType = OLAP_ABNORMAL_TERMINATION ; 
}

C_OLAPTuple_MemoryStream::~C_OLAPTuple_MemoryStream()
{
	//cerr << "destroying a memory stream" << endl ; 
	mem_stream_t*	isMemStream = NULL ;

	int32		count = 0 ; 
	while ( ( isMemStream = m_isInputList.pop() ) ) 
	{
		if ( isMemStream != m_isInputException && 
                     isMemStream != m_isOutputException )
		{
			count++ ; 
			//cerr << "deallocating mem ip " << isMemStream << endl ; 
			delete isMemStream ; 
		}
	}

	while ( ( isMemStream = m_isOutputList.pop() ) ) 
	{
		if ( isMemStream != m_isInputException && 
                     isMemStream != m_isOutputException )
		{
			count++ ; 
			//cerr << "deallocating mem op " << isMemStream << endl ; 
			delete isMemStream ; 
		}
	}

	//cerr << m_currInput << " " << m_currOutput << endl ; 

	if ( m_currInput && (m_currInput != m_isInputException && 
             m_currInput != m_isOutputException) )
	{
		//cerr << "deallocating input " << m_currInput << endl ; 
		delete m_currInput ; count++ ;  
	}

	if ( m_currOutput && (m_currOutput != m_isInputException && 
             m_currOutput != m_isOutputException) )
	{
		//cerr << "deallocating output " << m_currOutput << endl ; 
		delete m_currOutput ; count++ ; 
	}

	w_assert3( count == OLAP_MIN_PAGE_BUFFERS ) ; 

	delete m_isInputException ; 
	delete m_isOutputException ; 
}

w_rc_t C_OLAPTuple_MemoryStream::sendAPage( mem_stream_t* page, bool eos )
{
	W_DO( m_isStreamMutex.acquire() ) ;

	page->m_isMsgType = !eos ? OLAP_MORE_DATA : OLAP_END_OF_STREAM ; 
	m_isOutputList.append( page ) ; 

	m_isStreamMutex.release() ; m_isStreamCond.signal() ; 
	return RCOK ; 
}

w_rc_t C_OLAPTuple_MemoryStream::recvAPage( mem_stream_t*& page, bool& eos, bool& exception )
{
	W_DO( m_isStreamMutex.acquire() ) ;

	eos = false ; exception = false ; 

	// wait until a new page arrives on the stream
	while ( (page = m_isOutputList.pop()) == NULL )
		W_DO( m_isStreamCond.wait( m_isStreamMutex ) ) ;  

	if ( page->m_isMsgType == OLAP_END_OF_STREAM )
	{
		// release the page 
		page->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
		page->m_isPage->reset() ; 

		eos = true ; exception = false ; 
	} 
	else if ( page->m_isMsgType == OLAP_ABNORMAL_TERMINATION )
	{
		page->m_isPage->reset() ; 
		eos = false ; exception = true ; 
	}

	m_isStreamMutex.release() ; 
	return RCOK ; 
}

w_rc_t C_OLAPTuple_MemoryStream::releaseAPage( mem_stream_t* page )
{
	W_DO( m_isStreamMutex.acquire() ) ;

	page->m_isMsgType = OLAP_REQUEST_MORE_DATA ; 
	page->m_isPage->reset() ; 

	m_isInputList.append( page ) ; 

	m_isStreamMutex.release() ; m_isStreamCond.signal() ; 

	return RCOK ; 
}

w_rc_t C_OLAPTuple_MemoryStream::getAFreePage( mem_stream_t*& page, bool& exception )
{
	W_DO( m_isStreamMutex.acquire() ) ;

	// wait until a free page is available in the system 
	while ( (page = m_isInputList.pop()) == NULL )
		W_DO( m_isStreamCond.wait( m_isStreamMutex ) ) ;  

	exception = (page->m_isMsgType == OLAP_ABNORMAL_TERMINATION) ? true : false ;  
		
	m_isStreamMutex.release() ; 

	return RCOK ; 
} 

w_rc_t C_OLAPTuple_MemoryStream::read( C_OLAP_Tuple& aTuple, bool& eos, bool& exception )
{
	eos = exception = false ; 

	if ( !m_currOutput )
	{
		W_DO( recvAPage( m_currOutput, eos, exception ) ) ; 
		if ( eos || exception ) return RCOK ; 
	}

	bool eop ; 
	m_currOutput->m_isPage->read( aTuple, eop ) ;

	if ( eop )
	{
		W_DO( releaseAPage( m_currOutput ) ) ; 
		W_DO( recvAPage( m_currOutput, eos, exception ) ) ; 
		
		if ( !eos && !exception ) 	
			m_currOutput->m_isPage->read( aTuple, eop )  ; 	
	}

	return RCOK ; 
}

w_rc_t C_OLAPTuple_MemoryStream::write( C_OLAP_Tuple& aTuple, bool& exception )
{
	exception = false ; 

	if ( !m_currInput )
	{
		W_DO( getAFreePage( m_currInput, exception ) ) ; 
		if ( exception ) return RCOK ;  
	}

	bool page_full ; 
	m_currInput->m_isPage->write( aTuple, page_full ) ;

	if ( page_full ) 
	{
		W_DO( sendAPage( m_currInput ) ) ; 
		W_DO( getAFreePage( m_currInput, exception ) ) ; 

		if ( !exception )
			m_currInput->m_isPage->write( aTuple, page_full )  ;
	}

	return RCOK ; 
}


w_rc_t C_OLAPTuple_MemoryStream::read( C_OLAP_Page*& page, bool& eos, bool& exception )
{
	eos = exception = false ; 
	mem_stream_t*	new_page = NULL ; 

	W_DO( recvAPage( new_page, eos, exception ) ) ; 
	if ( eos || exception ) 
	{
		W_DO( releaseAPage( new_page ) ) ; 
		return RCOK ; 
	}

	page = new_page->m_isPage ; 

	new_page->m_isPage = new C_OLAP_Page ; 
	if ( !new_page->m_isPage ) W_FATAL(BORG_OUTOFMEMORY) ; 

	W_DO( releaseAPage( new_page ) ) ; 

        return RCOK ;
}

w_rc_t C_OLAPTuple_MemoryStream::write( C_OLAP_Page* page, bool& exception )
{
	exception = false ; 
	mem_stream_t*	free_page = NULL ; 

	W_DO( getAFreePage( free_page, exception ) ) ; 
	if ( exception ) 
	{
		W_DO( releaseAPage( free_page ) ) ; 
		return RCOK ; 
	}

	if ( free_page->m_isPage ) 
		delete free_page->m_isPage ; 
	else
		w_assert3(0) ; 

	free_page->m_isPage = page ; 
	W_DO( sendAPage( free_page ) ) ; 

        return RCOK ;
}

w_rc_t C_OLAPTuple_MemoryStream::SendEndOfStream()
{
	if ( m_currInput )
		W_DO( sendAPage( m_currInput ) ) ; 

	bool exception ; 
	W_DO( getAFreePage( m_currInput, exception ) ) ; 

	if ( exception ) return RCOK ; 
 
	W_DO( sendAPage( m_currInput, true ) ) ; 

	m_currInput = NULL ; 
	return RCOK ; 
}

w_rc_t C_OLAPTuple_MemoryStream::SendExceptionDownStream()
{
	//cerr << "Sending an downstream exception" << endl ; 

	W_DO( m_isStreamMutex.acquire() ) ;

	m_isInputException->m_isMsgType = OLAP_ABNORMAL_TERMINATION ;

	// add the abnormal message at the top of the list 
	m_isInputList.push(m_isInputException) ; 

	m_isStreamMutex.release() ; m_isStreamCond.signal() ; 

	return RCOK ; 
}

w_rc_t C_OLAPTuple_MemoryStream::SendExceptionUpStream()
{
	//cerr << "Sending an upstream exception" << endl ; 

	W_DO( m_isStreamMutex.acquire() ) ;

	m_isOutputException->m_isMsgType = OLAP_ABNORMAL_TERMINATION ;

	// add the abnormal message at the top of the list 
	m_isOutputList.push(m_isOutputException) ; 

	m_isStreamMutex.release() ; m_isStreamCond.signal() ; 

	return RCOK ; 
}


#ifdef __GNUG__
template class w_list_t<C_OLAPTuple_MemoryStream::mem_stream_t>;
template class w_list_i<C_OLAPTuple_MemoryStream::mem_stream_t>;
#endif

