/*
Copyright (c) 1997-1998
Borg Team
University of Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE BORG TEAM AT THE UNIVERSITY OF WISCONSIN -- MADISON ALLOWS FREE
USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION.  WE DISCLAIM ANY
LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
THE USE OF THIS SOFTWARE.

The Borg Team consists of Prasad Deshpande, Karthikeyan Ramasamy,
Amit Shukla and Jeffrey Naughton
*/

////////////////////////////////////////////////////////////////////
//
// olap_tuple.h
//
// Declares tuple related classes:
// C_OLAP_TupleInfo 	-- stores information about tuple attributes
// C_OLAP_Tuple  	-- the actual tuple class
// C_OLAP_TupleIterator -- used to iterate over attributes in a tuple
//
// Author: Prasad M. Deshpande
//
/////////////////////////////////////////////////////////////////////



#ifndef _OLAP_TUPLE_H
#define _OLAP_TUPLE_H

#include "olap_types.h"
#include "olap_basics.h"
#include "borgtypes.h"
#include "w_dynarray.h"
#include "olap_comm_data.h"

class OlapType;
enum BorgTypes;

class C_OLAPAttrDesc : public C_OLAPComm_Data
{
public:
    OlapType 	*m_type;     // type of the attribute
    u_int32     m_offset;    // offset into the tuple


    // Functions for shipping
    void Linearize( C_OLAPComm_DataLinearizer &dLin );

    void DeLinearize( C_OLAPComm_DataDelinearizer &deLin );

};

class C_OLAP_TupleInfo : public C_OLAPComm_Data
{
public:
    C_OLAP_TupleInfo() : m_numAttr(0), m_attrDescs(10), m_fixedLen(0) {
    }

    ~C_OLAP_TupleInfo() {
    }

    void SetNumAttrs(u_int32 nAttr) {
	m_numAttr = nAttr;
	m_attrDescs.set_size( m_numAttr );
    }


    void SetAttrDesc(u_int32 attrNum, BorgTypes attrTypeid) {
	borg_assert3(attrNum < m_numAttr);
	m_attrDescs[attrNum].m_type = &OlapType::BasicOlapTypes[attrTypeid];
    }

    void EndAttrDesc();

    u_int32 GetFixedLen() {
	return m_fixedLen;
    }

    C_OLAPAttrDesc &GetAttrDesc(u_int32 attrNum) {
	borg_assert3(attrNum < m_numAttr);
	return m_attrDescs[attrNum];
    }

    C_OLAPAttrDesc &operator[](u_int32 attrNum) {
	borg_assert3(attrNum < m_numAttr);
	return m_attrDescs[attrNum];
    }

    u_int32 GetNumAttrs() {
	return m_numAttr;
    }

    C_OLAP_TupleInfo &operator=(C_OLAP_TupleInfo &rhs);

    // Functions for shipping
    void Linearize( C_OLAPComm_DataLinearizer &dLin );

    void DeLinearize( C_OLAPComm_DataDelinearizer &deLin );
    
private:
    u_int32 			m_numAttr;
    dynarray<C_OLAPAttrDesc>  	m_attrDescs;
    u_int32			m_fixedLen;
};    

class C_OLAP_Tuple
{
public:

    static const int32 TupInitSize = 256;

    C_OLAP_Tuple() : m_tupBuf(NULL), m_varOffset(0), 
	m_allocFlag(false), m_size(0), m_fixedSize(0), m_tupInfo(NULL) {
    }

    ~C_OLAP_Tuple() {
	releaseBuf();
    }

    void Init(C_OLAP_TupleInfo &tupInfo) {
	m_tupInfo = &tupInfo;
	m_fixedSize = m_varOffset = tupInfo.GetFixedLen();
    }

    void AllocBuf() {
	alloc( m_fixedSize );
    }

    void SetFixedLen(u_int32 fixedSize) {
	m_fixedSize = m_varOffset = fixedSize;
    }

    u_int32 GetFixedLen() {
	return m_fixedSize ;
    }

    void Reset() {
	m_varOffset = m_fixedSize;
    }

    void *GetTuple() const {
	return (void *)m_tupBuf;
    }

    u_int32 GetLength() const {
	return m_size;
    }

    void FillAttr(C_OLAPAttrDesc &attr, void *buf, int32 length=-1);

    void *operator[](u_int32 attrNum) {
	if (!m_tupInfo || (attrNum >= m_tupInfo->GetNumAttrs()))
	    return NULL;
	u_int32 len;
	return GetAttr((*m_tupInfo)[attrNum], len);
    }

    void *GetAttr(u_int32 attrNum, u_int32 &length) {
	if (!m_tupInfo || (attrNum >= m_tupInfo->GetNumAttrs()))
	    return NULL;
	return GetAttr((*m_tupInfo)[attrNum],length);
    }	

    void print(ostream &str);

private:

    char 	*m_tupBuf;   // tuple buffer
    u_int32 	m_varOffset; // offset for variable part of the tuple
    bool 	m_allocFlag; // flag for memory allocation
    u_int32 	m_size;      // size of buf
    u_int32 	m_fixedSize; // fixed size of the tuple
    C_OLAP_TupleInfo *m_tupInfo; // pointer to tuple info

    void releaseBuf() {
	if (m_allocFlag)
	    DelA(m_tupBuf);
	m_allocFlag = false;
	m_tupBuf = NULL;
	m_varOffset = 0;
	m_size = 0;
    }	

    void SetTuple(void *buf, u_int32 size) {
	releaseBuf();
	m_tupBuf = (char *)buf;
	m_size = size;
    }

    void alloc(u_int32 len);

    void *GetAttr(C_OLAPAttrDesc &attr, u_int32 &length) const;

    friend class C_OLAP_Page ;
    friend class C_OLAP_TupleIterator ;
    friend class page ;
};


class C_OLAP_TupleIterator
{
public:
    
    C_OLAP_TupleIterator()          // default constructor - init has to be called
    { }

    C_OLAP_TupleIterator(C_OLAP_TupleInfo &tupInfo, C_OLAP_Tuple &tuple) {
	init(tupInfo, tuple);
    }

    void init(C_OLAP_TupleInfo &tupInfo, C_OLAP_Tuple &tuple) {
	m_tupInfo = &tupInfo;
	m_tuple = &tuple;
	m_currAttr = 0;
    }

    C_OLAP_TupleIterator &begin() {
	m_currAttr = 0;
	return *this;
    }

    void *operator*() {
	if (m_currAttr == m_tupInfo->GetNumAttrs())
	    return NULL;
	u_int32 len;
	return m_tuple->GetAttr((*m_tupInfo)[m_currAttr], len);
    }

    C_OLAP_TupleIterator &operator++(int) {
	if (m_currAttr < m_tupInfo->GetNumAttrs())
	    m_currAttr++;
	return *this;
    }

    void *operator[](u_int32 attrNum) {
	if (attrNum >= m_tupInfo->GetNumAttrs())
	    return NULL;
	u_int32 len;
	return m_tuple->GetAttr((*m_tupInfo)[attrNum], len);
    }

    void *GetAttr(u_int32 attrNum, u_int32 &length) {
	if (attrNum >= m_tupInfo->GetNumAttrs())
	    return NULL;
	return m_tuple->GetAttr((*m_tupInfo)[attrNum],length);
    }	

private:
    C_OLAP_TupleInfo	*m_tupInfo;
    C_OLAP_Tuple  	*m_tuple;
    u_int32  		m_currAttr;
};

#endif
